/********************************************************************
 File: main.c

 Description:
 This file contains the main logic of the USBStickLogger project.
 USBStickLogger records serial data received from the UART on the
 connected USB thumb drive.

 Authors and Copyright:
 (c) 2011, Thomas Fischl <tfischl@gmx.de>

 Device: PIC24FJ64GB002
 Compiler: C30

 License:
 This file is open source. You can use it or parts of it in own
 open source projects. For closed or commercial projects you have to
 contact the authors listed above to get the permission for using
 this source code.

 This firmware uses libraries provided by Microchip. Please read the
 software licence agreement in the particular file for further
 information if you intend to use parts of this library in your own
 programs.
 
 Change History:
  Rev   Date        Description
  1.0   2011-04-10  Initial release
  1.1   2011-07-23  Fixed serialbuffer fill handling,
                    changed baudrate generator to high speed
  1.2   2011-08-06  Improved LED signaling

********************************************************************/
#include "USB/usb.h"
#include "USB/usb_host_msd.h"
#include "USB/usb_host_msd_scsi.h"
#include "MDD File System/FSIO.h"

// configuration bits
_CONFIG1(WDTPS_PS1 & FWPSA_PR32 & WINDIS_OFF & FWDTEN_OFF & ICS_PGx1 & GWRP_OFF & GCP_OFF & JTAGEN_OFF)
_CONFIG2(POSCMOD_HS & I2C1SEL_PRI & IOL1WAY_OFF & OSCIOFNC_ON & FCKSM_CSDCMD & FNOSC_PRIPLL & PLL96MHZ_OFF & PLLDIV_DIV3 & IESO_ON)
_CONFIG3(WPFP_WPFP0 & SOSCSEL_SOSC & WUTSEL_LEG & WPDIS_WPDIS & WPCFG_WPCFGDIS & WPEND_WPENDMEM)
_CONFIG4(DSWDTPS_DSWDTPS3 & DSWDTOSC_LPRC & RTCOSC_SOSC & DSBOREN_OFF & DSWDTEN_OFF)


#define LED_PWM_STEP 300
#define LED_FADE_OFF 0
#define LED_FADE_ON 1
#define LED_PWMVALUE_ON 0xffff
#define LED_PWMVALUE_OFF 0

volatile unsigned char led_fade = 1;
volatile unsigned short led_pwmvalue = 0;

void led_set(unsigned char fade, unsigned short pwmvalue) {
    led_fade = fade;
    led_pwmvalue = pwmvalue;
    OC1R = led_pwmvalue;
}

void led_setValue(unsigned short pwmvalue) {
    led_pwmvalue = pwmvalue;
    OC1R = led_pwmvalue;
}



#define SWITCH (!PORTBbits.RB5)

#define SERIALBUFFER_SIZE 64
#define SERIALBUFFER_COUNT 2
#define SERIALBUFFER_OWNER_SERIAL 0
#define SERIALBUFFER_OWNER_FS 1

struct serialbuffer_s {
  volatile BYTE owner;
  volatile BYTE position;
  BYTE data[SERIALBUFFER_SIZE];
} serialbuffer[SERIALBUFFER_COUNT];

unsigned char serialbuffer_serial_current;
unsigned char serialbuffer_fs_current;

void serialbuffer_init() {
	unsigned char i;
	for (i = 0; i < SERIALBUFFER_COUNT; i++) {
		serialbuffer[i].owner = SERIALBUFFER_OWNER_SERIAL;
		serialbuffer[i].position = 0;
	}
	serialbuffer_serial_current = 0;
 	serialbuffer_fs_current = 0;
}

void serialbuffer_putch(unsigned char ch) {

    if (serialbuffer[serialbuffer_serial_current].owner != SERIALBUFFER_OWNER_SERIAL) {
        // Overflow; signal it
		led_set(LED_FADE_OFF, LED_PWMVALUE_ON);
        return;
    }

	serialbuffer[serialbuffer_serial_current].data[serialbuffer[serialbuffer_serial_current].position] = ch;
	serialbuffer[serialbuffer_serial_current].position++;
	if (serialbuffer[serialbuffer_serial_current].position == SERIALBUFFER_SIZE) {
		serialbuffer[serialbuffer_serial_current].owner = SERIALBUFFER_OWNER_FS;
		serialbuffer_serial_current ++;
		if (serialbuffer_serial_current == SERIALBUFFER_COUNT) {
			serialbuffer_serial_current = 0;
		}
	}
}



size_t freadln(char *ptr, size_t maxsize, FSFILE *stream) {
	size_t count = 0;
	WORD charsRead;

	while (!FSfeof(stream)) {
		charsRead = FSfread(&ptr[count], 1, 1, stream);
		if (!charsRead) break;
		if (ptr[count] == 10) break;
		if (ptr[count] != 13)
			count += charsRead;
		if (count >= maxsize) {
			count = maxsize - 1;
			break;
		}
	}
	ptr[count] = 0;
	return count;
}

BOOL parseLine(char *line, char* key, char* value) {
	WORD pos = 0;
	WORD pos_target = 0;
	key[0] = 0;
	value[0] = 0;

	while (1) {
		char ch = line[pos];
		pos++;
		if (ch == 0) return FALSE;
		if (ch == ':') break;
		
		if ((ch >= 'a') && (ch <= 'z')) {
			ch = 'A' + (ch - 'a');
		}

		if (((ch >= 'A') && (ch <= 'Z')) ||
            ((ch >= '0') && (ch <= '9')) ||
            (ch == '_')) {
			key[pos_target] = ch;
			pos_target++;
			key[pos_target] = 0;
		}
	}
	pos_target = 0;
	while (1) {
		char ch = line[pos];
		pos++;
		if (ch == 0) break;
		
		if ((ch >= 'a') && (ch <= 'z')) {
			ch = 'A' + (ch - 'a');
		}

		if (((ch >= 'A') && (ch <= 'Z')) ||
            ((ch >= '0') && (ch <= '9')) ||
            (ch == '_')) {
			value[pos_target] = ch;
			pos_target++;
			value[pos_target] = 0;
		}
	}

	return TRUE;
}

void lockIO(){
	asm volatile ("mov #OSCCON,w1 \n"
                "mov #0x46, w2 \n"
                "mov #0x57, w3 \n"
                "mov.b w2,[w1] \n"
                "mov.b w3,[w1] \n"
                "bset OSCCON, #6");
}
 
void unlockIO(){
	asm volatile ("mov #OSCCON,w1 \n"
                "mov #0x46, w2 \n"
                "mov #0x57, w3 \n"
                "mov.b w2,[w1] \n"
                "mov.b w3,[w1] \n"
                "bclr OSCCON, #6");
}

void uart_setBaudrate(long baudrate) {
    U1MODEbits.BRGH = 1;
	U1BRG = GetInstructionClock() / 4 / baudrate - 1;
}

volatile BOOL deviceAttached;




int main(void)
{

	// power up the PLL
    {
        unsigned int pll_startup_counter = 600;
        CLKDIVbits.PLLEN = 1;
        while (pll_startup_counter--);
    }

	// set all analog pins to digital
	AD1PCFG = 0xFFFF;   

	// map peripheral pins
	unlockIO();
	RPINR18bits.U1RXR = 9; // map RX to RP9 (pin 17)
    RPOR2bits.RP4R = 3; // map TX to RP4 (pin 11)
    RPOR7bits.RP15R = 18; // map OC1 to RP15 (led)
    lockIO();

	// init UART1
	serialbuffer_init();
	TRISBbits.TRISB4 = 0; // set TX pin to output
	U1MODE = 0; // data, parity, stop bits
	uart_setBaudrate(9600);
	U1MODEbits.UARTEN = 1; // enable uart
    U1STAbits.UTXEN = 1;
    IFS0bits.U1RXIF = 0;
	IEC0bits.U1RXIE = 1; // enable interrupt
   

	// init LED
	TRISBbits.TRISB15 = 0;
    led_set(LED_FADE_ON, LED_PWMVALUE_OFF);

	// init switch
	CNPU2bits.CN27PUE = 1; // switch pullup on

    // init timer
    T1CON = 0x8000; // timer1 on, 1:1 prescaler 

    // init output compare / pwm
    OC1RS = 0xffff;
    OC1CON2 = 0x001F;
    OC1CON1 = 0x1C06; // system clock

    // init usb stack
    deviceAttached = FALSE;
    USBInitialize(0);

	FSFILE * myFile = NULL;
	BOOL logging = FALSE;

unsigned short blinktest = 0;
    while(1)
    {
       if (IFS0bits.T1IF) {
          IFS0bits.T1IF = 0;
          if (led_fade) {
             if (led_pwmvalue > LED_PWM_STEP) led_pwmvalue -= LED_PWM_STEP;
             else led_pwmvalue = 0;
          } 
          OC1R = led_pwmvalue;
       }

		// USB stack process function
        USBTasks();

		// check if USB thumb drive was plugged in
		if ((deviceAttached != TRUE) && USBHostMSDSCSIMediaDetect()) {

			deviceAttached = TRUE;			

			if (FSInit())
            {

				// read config
				myFile = FSfopen("config.txt", "r");
				if (myFile != NULL) {
					char line[100], key[50], value[50];
					while (!FSfeof(myFile)) {
						freadln(line, 100, myFile);
						if (parseLine(line, key, value)) {
							if (!strncmp(key, "COM_BAUDRATE", 12) && (strlen(key) == 12)) {
								long baudrate = atol(value);
								if (baudrate >= 300) {
									uart_setBaudrate(baudrate);
								}
								
							}
						}
					}
					FSfclose(myFile);
				}

				// open logging file
                myFile = FSfopen("logging.txt", "a");
				if (myFile != NULL)
					logging = TRUE;
			}
                
		}

		if (logging && !deviceAttached) {
			// unclean detached!
			logging = FALSE;
		}

		if (logging && SWITCH) {
			led_set(LED_FADE_OFF, LED_PWMVALUE_ON);

			// write pending buffer
			FSfwrite(serialbuffer[serialbuffer_fs_current].data, 1, serialbuffer[serialbuffer_fs_current].position, myFile);
			serialbuffer[serialbuffer_fs_current].position = 0;

			FSfclose(myFile);
			logging = FALSE;
			led_set(LED_FADE_ON, LED_PWMVALUE_OFF);
		}

		if (serialbuffer[serialbuffer_fs_current].owner == SERIALBUFFER_OWNER_FS) {

			if (logging) {
				FSfwrite(serialbuffer[serialbuffer_fs_current].data, 1, serialbuffer[serialbuffer_fs_current].position, myFile);
			}
			serialbuffer[serialbuffer_fs_current].position = 0;
			serialbuffer[serialbuffer_fs_current].owner = SERIALBUFFER_OWNER_SERIAL;
			serialbuffer_fs_current ++;
			if (serialbuffer_fs_current == SERIALBUFFER_COUNT) {
				serialbuffer_fs_current = 0;
			}
		}

    }
    return 0;
}


/****************************************************************************
  Function:
    BOOL USB_ApplicationEventHandler( BYTE address, USB_EVENT event,
                void *data, DWORD size )

  Summary:
    This is the application event handler.  It is called when the stack has
    an event that needs to be handled by the application layer rather than
    by the client driver.

  Description:
    This is the application event handler.  It is called when the stack has
    an event that needs to be handled by the application layer rather than
    by the client driver.  If the application is able to handle the event, it
    returns TRUE.  Otherwise, it returns FALSE.

  Precondition:
    None

  Parameters:
    BYTE address    - Address of device where event occurred
    USB_EVENT event - Identifies the event that occured
    void *data      - Pointer to event-specific data
    DWORD size      - Size of the event-specific data

  Return Values:
    TRUE    - The event was handled
    FALSE   - The event was not handled

  Remarks:
    The application may also implement an event handling routine if it
    requires knowledge of events.  To do so, it must implement a routine that
    matches this function signature and define the USB_HOST_APP_EVENT_HANDLER
    macro as the name of that function.
  ***************************************************************************/

BOOL USB_ApplicationEventHandler( BYTE address, USB_EVENT event, void *data, DWORD size )
{
    switch( event )
    {
        case EVENT_VBUS_REQUEST_POWER:
            // The data pointer points to a byte that represents the amount of power
            // requested in mA, divided by two.  If the device wants too much power,
            // we reject it.
            return TRUE;

        case EVENT_VBUS_RELEASE_POWER:
            // Turn off Vbus power.
            // The PIC24F with the Explorer 16 cannot turn off Vbus through software.

            //This means that the device was removed
            deviceAttached = FALSE;
            return TRUE;
            break;

        case EVENT_HUB_ATTACH:
            return TRUE;
            break;

        case EVENT_UNSUPPORTED_DEVICE:
            return TRUE;
            break;

        case EVENT_CANNOT_ENUMERATE:
            //UART2PrintString( "\r\n***** USB Error - cannot enumerate device *****\r\n" );
            return TRUE;
            break;

        case EVENT_CLIENT_INIT_ERROR:
            //UART2PrintString( "\r\n***** USB Error - client driver initialization error *****\r\n" );
            return TRUE;
            break;

        case EVENT_OUT_OF_MEMORY:
            //UART2PrintString( "\r\n***** USB Error - out of heap memory *****\r\n" );
            return TRUE;
            break;

        case EVENT_UNSPECIFIED_ERROR:   // This should never be generated.
            //UART2PrintString( "\r\n***** USB Error - unspecified *****\r\n" );
            return TRUE;
            break;

        default:
            break;
    }

    return FALSE;
}




void __attribute__((interrupt, auto_psv)) _U1RXInterrupt(void) {
	unsigned char ch = U1RXREG;
	IFS0bits.U1RXIF = 0;

	 serialbuffer_putch(ch);
     led_setValue(LED_PWMVALUE_ON);
}

